from __future__ import annotations

from typing import List

import chex
import numpy as np
from gymnasium.core import RenderFrame

from moojoco.mjcf.arena import MJCFArena
from moojoco.environment.base import (
    BaseEnvState,
    BaseEnvironment,
    MuJoCoEnvironmentConfiguration,
    SpaceType,
)
from moojoco.environment.mjc_env import MJCEnv
from moojoco.environment.mjx_env import MJXEnv
from moojoco.mjcf.morphology import MJCFMorphology


class DualMuJoCoEnvironment(BaseEnvironment):
    MJC_ENV_CLASS: type[MJCEnv]
    MJX_ENV_CLASS: type[MJXEnv]

    def __init__(self, env: MJCEnv | MJXEnv, backend: str) -> None:
        super().__init__(configuration=env.environment_configuration)
        self.backend = backend
        self._env = env

    @classmethod
    def from_morphology_and_arena(
        cls,
        morphology: MJCFMorphology,
        arena: MJCFArena,
        configuration: MuJoCoEnvironmentConfiguration,
        backend: str,
    ) -> DualMuJoCoEnvironment:
        assert backend in [
            "MJC",
            "MJX",
        ], f"Backend must either be 'MJC' or 'MJX'. {backend} was given."
        if backend == "MJC":
            env_class = cls.MJC_ENV_CLASS
        else:
            env_class = cls.MJX_ENV_CLASS
        env = env_class.from_morphology_and_arena(
            morphology=morphology, arena=arena, configuration=configuration
        )
        return cls(env=env, backend=backend)

    @property
    def action_space(self) -> SpaceType:
        return self._env.action_space

    @property
    def actuators(self) -> List[str]:
        return self._env.actuators

    @property
    def observation_space(self) -> SpaceType:
        return self._env.observation_space

    def step(
        self, state: BaseEnvState, action: chex.Array, *args, **kwargs
    ) -> BaseEnvState:
        return self._env.step(state=state, action=action, *args, **kwargs)

    def reset(
        self, rng: np.random.RandomState | chex.PRNGKey, *args, **kwargs
    ) -> BaseEnvState:
        return self._env.reset(rng=rng, *args, **kwargs)

    def render(self, state: BaseEnvState) -> List[RenderFrame] | None:
        return self._env.render(state=state)

    def close(self) -> None:
        return self._env.close()



from __future__ import annotations

import abc
import copy
from abc import ABC
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Callable, Dict, List, Tuple

import gymnasium
import mujoco
import numpy as np
from flax import struct
from gymnasium.core import ActType, RenderFrame
from gymnasium.vector.utils import batch_space

from moojoco.environment.base import (
    BaseEnvState,
    BaseEnvironment,
    BaseMuJoCoEnvironment,
    BaseObservable,
    MuJoCoEnvironmentConfiguration,
)
from moojoco.mjcf.arena import MJCFArena
from moojoco.mjcf.morphology import MJCFMorphology


@struct.dataclass
class MJCEnvState(BaseEnvState):
    observations: Dict[str, np.ndarray]
    rng: np.random.RandomState


class MJCObservable(BaseObservable):
    def __init__(
        self,
        name: str,
        low: np.ndarray,
        high: np.ndarray,
        retriever: Callable[[MJCEnvState], np.ndarray],
    ) -> None:
        super().__init__(name=name, low=low, high=high, retriever=retriever)

    def __call__(self, state: MJCEnvState) -> np.ndarray:
        return super().__call__(state=state)


class MJCEnv(BaseMuJoCoEnvironment, ABC):
    def __init__(
        self,
        mjcf_str: str,
        mjcf_assets: Dict[str, Any],
        configuration: MuJoCoEnvironmentConfiguration,
    ) -> None:
        super().__init__(
            mjcf_str=mjcf_str, mjcf_assets=mjcf_assets, configuration=configuration
        )

    @classmethod
    def from_morphology_and_arena(
        cls,
        morphology: MJCFMorphology,
        arena: MJCFArena,
        configuration: MuJoCoEnvironmentConfiguration,
    ) -> MJCEnv:
        return super().from_morphology_and_arena(
            morphology=morphology, arena=arena, configuration=configuration
        )

    def _get_mj_models_and_datas_to_render(
        self, state: MJCEnvState
    ) -> Tuple[List[mujoco.MjModel], List[mujoco.MjData]]:
        return [state.mj_model], [state.mj_data]

    @property
    def observables(self) -> List[MJCObservable]:
        return self._observables

    def _create_observation_space(self) -> gymnasium.spaces.Dict:
        observation_space = dict()
        for observable in self.observables:
            observation_space[observable.name] = gymnasium.spaces.Box(
                low=observable.low, high=observable.high, shape=observable.shape
            )
        return gymnasium.spaces.Dict(observation_space)

    def _create_action_space(self) -> gymnasium.Space:
        bounds = self.frozen_mj_model.actuator_ctrlrange.copy().astype(np.float32)
        low, high = bounds.T
        action_space = gymnasium.spaces.Box(low=low, high=high, dtype=np.float32)
        return action_space

    def _update_observations(self, state: MJCEnvState) -> MJCEnvState:
        observations = dict()
        for observable in self.observables:
            observations[observable.name] = observable(state=state)
        # noinspection PyUnresolvedReferences
        return state.replace(observations=observations)

    def _update_simulation(self, state: MJCEnvState, ctrl: ActType) -> MJCEnvState:
        mj_data = copy.deepcopy(state.mj_data)
        mj_data.ctrl[:] = ctrl
        mujoco.mj_step(
            m=state.mj_model,
            d=mj_data,
            nstep=self.environment_configuration.num_physics_steps_per_control_step,
        )
        # As of MuJoCo 2.0, force-related quantities like cacc are not computed
        # unless there's a force sensor in the model.
        # See https://github.com/openai/gym/issues/1541
        mujoco.mj_rnePostConstraint(state.mj_model, mj_data)

        # noinspection PyUnresolvedReferences
        return state.replace(mj_data=mj_data)

    def _prepare_reset(self) -> Tuple[mujoco.MjModel, mujoco.MjData]:
        model = copy.deepcopy(self.frozen_mj_model)
        data = copy.deepcopy(self.frozen_mj_data)
        return model, data

    def _finish_reset(
        self,
        models_and_datas: Tuple[mujoco.MjModel, mujoco.MjData],
        rng: np.random.RandomState,
    ) -> MJCEnvState:
        mj_model, mj_data = models_and_datas
        mujoco.mj_forward(m=mj_model, d=mj_data)
        state = MJCEnvState(
            mj_model=mj_model,
            mj_data=mj_data,
            observations={},
            reward=0,
            terminated=False,
            truncated=False,
            info={},
            rng=rng,
        )
        state = self._update_observations(state=state)
        state = self._update_info(state=state)
        return state

    @abc.abstractmethod
    def _create_observables(self) -> List[MJCObservable]:
        raise NotImplementedError

    @abc.abstractmethod
    def reset(self, rng: np.random.RandomState, *args, **kwargs) -> MJCEnvState:
        raise NotImplementedError

    @abc.abstractmethod
    def _update_reward(
        self, state: MJCEnvState, previous_state: MJCEnvState
    ) -> MJCEnvState:
        raise NotImplementedError

    @abc.abstractmethod
    def _update_terminated(self, state: MJCEnvState) -> MJCEnvState:
        raise NotImplementedError

    @abc.abstractmethod
    def _update_truncated(self, state: MJCEnvState) -> MJCEnvState:
        raise NotImplementedError

    @abc.abstractmethod
    def _update_info(self, state: MJCEnvState) -> MJCEnvState:
        raise NotImplementedError


@struct.dataclass
class VectorMJCEnvState(MJCEnvState):
    mj_model: List[mujoco.MjModel] = struct.field(pytree_node=False)
    mj_data: List[mujoco.MjData] = struct.field(pytree_node=False)
    observations: Dict[str, np.ndarray]
    reward: np.ndarray
    terminated: np.ndarray
    truncated: np.ndarray
    info: Dict[str, Any]
    rng: List[np.random.RandomState]


class ThreadedVectorMJCEnvWrapper(BaseEnvironment):
    def __init__(
        self, create_env_fn: Callable[[], BaseEnvironment], num_environments: int
    ) -> None:
        self._create_env_fn = create_env_fn
        self._num_environments = num_environments
        dummy_env = create_env_fn()
        self._single_action_space = dummy_env.action_space
        self._action_space = batch_space(
            self._single_action_space, self._num_environments
        )
        self._observation_space = batch_space(
            dummy_env.observation_space, self._num_environments
        )
        self._actuators = dummy_env.actuators
        super().__init__(configuration=dummy_env.environment_configuration)
        dummy_env.close()

        self._pool = ThreadPoolExecutor(max_workers=num_environments)
        self._envs = self._create_envs()
        self._states = None

    def _create_envs(self) -> List[MJCEnv]:
        environments = list(
            self._pool.map(
                lambda _: self._create_env_fn(), range(self._num_environments)
            )
        )
        return environments

    @property
    def _merged_states(self) -> VectorMJCEnvState:
        mj_models = []
        mj_datas = []
        observations = defaultdict(list)
        reward = []
        terminated = []
        truncated = []
        info = defaultdict(list)
        rng = []
        for env_id, state in enumerate(self._states):
            mj_models.append(state.mj_model)
            mj_datas.append(state.mj_data)
            for k, o in state.observations.items():
                observations[k].append(o)
            reward.append(state.reward)
            terminated.append(state.terminated)
            truncated.append(state.truncated)
            for k, o in state.info.items():
                info[k].append(o)
            rng.append(state.rng)

        observations = {k: np.stack(v) for k, v in observations.items()}
        info = {k: np.array(v) for k, v in info.items()}

        return VectorMJCEnvState(
            mj_model=mj_models,
            mj_data=mj_datas,
            observations=observations,
            reward=np.array(reward),
            terminated=np.array(terminated),
            truncated=np.array(truncated),
            info=info,
            rng=rng,
        )

    @property
    def single_action_space(self) -> gymnasium.spaces.Space:
        return self._single_action_space

    @property
    def action_space(self) -> gymnasium.spaces.Space:
        return self._action_space

    @property
    def actuators(self) -> List[str]:
        return self._actuators

    @property
    def observation_space(self) -> gymnasium.spaces.Space:
        return self._observation_space

    def step(
        self, state: VectorMJCEnvState, action: np.ndarray, *args, **kwargs
    ) -> VectorMJCEnvState:
        self._states = list(
            self._pool.map(
                lambda env, ste, act: env.step(state=ste, action=act, *args, **kwargs),
                self._envs,
                self._states,
                action,
            )
        )
        return self._merged_states

    def reset(
        self, rng: List[np.random.RandomState], *args, **kwargs
    ) -> VectorMJCEnvState:
        self._states = list(
            self._pool.map(
                lambda env, sub_rng: env.reset(sub_rng, *args, **kwargs),
                self._envs,
                rng,
            )
        )

        return self._merged_states

    def render(self, state: VectorMJCEnvState) -> List[RenderFrame] | None:
        if self.environment_configuration.render_mode == "human":
            # Only render first env; Has to be in main thread
            return self._envs[0].render(state=self._states[0])

        frames = []
        for env, state in zip(self._envs, self._states):
            frames += env.render(state=state)
        return frames

    def close(self) -> None:
        self._pool.shutdown()
        for env in self._envs:
            env.close()



import abc
from typing import List, Tuple

import chex
import numpy as np
from gymnasium.core import RenderFrame

from moojoco.environment.base import BaseEnvState, BaseEnvironment, SpaceType


class PostInitCaller(type):
    def __call__(cls, *args, **kwargs):
        obj = type.__call__(cls, *args, **kwargs)
        obj.__post__init__()
        return obj


class CombinedMeta(abc.ABCMeta, PostInitCaller):
    pass


class EnvironmentWrapper(BaseEnvironment, metaclass=CombinedMeta):
    def __init__(self, env: BaseEnvironment) -> None:
        super().__init__(configuration=env.environment_configuration)
        self._env = env

        self._observation_space: SpaceType | None = None
        self._action_space: SpaceType | None = None

    def __post__init__(self) -> None:
        # Make sure observation space and action space are initialised after creation
        # noinspection PyStatementEffect
        self.observation_space
        # noinspection PyStatementEffect
        self.action_space

    @property
    def action_space(self) -> SpaceType:
        if self._action_space is not None:
            return self._action_space
        return self._env.action_space

    @property
    def actuators(self) -> List[str]:
        return self._env.actuators

    @property
    def observation_space(self) -> SpaceType:
        if self._observation_space is not None:
            return self._observation_space
        return self._env.observation_space

    def step(
        self, state: BaseEnvState, action: chex.Array, *args, **kwargs
    ) -> BaseEnvState:
        return self._env.step(state=state, action=action, *args, **kwargs)

    def reset(
        self, rng: np.random.RandomState | chex.PRNGKey, *args, **kwargs
    ) -> BaseEnvState:
        return self._env.reset(rng=rng, *args, **kwargs)

    def render(self, state: BaseEnvState) -> List[RenderFrame] | None:
        return self._env.render(state=state)

    def close(self) -> None:
        return self._env.close()


class TransformObservationEnvWrapper(EnvironmentWrapper, abc.ABC):
    def __init__(self, env: BaseEnvironment) -> None:
        super().__init__(env=env)

    @property
    @abc.abstractmethod
    def observation_space(self) -> SpaceType:
        raise NotImplementedError

    @abc.abstractmethod
    def _transform_observations(self, state: BaseEnvState) -> BaseEnvState:
        raise NotImplementedError

    def step(
        self, state: BaseEnvState, action: chex.Array, *args, **kwargs
    ) -> BaseEnvState:
        state = self._env.step(state=state, action=action, *args, **kwargs)
        state = self._transform_observations(state=state)
        return state

    def reset(
        self, rng: np.random.RandomState | chex.PRNGKey, *args, **kwargs
    ) -> BaseEnvState:
        state = self._env.reset(rng=rng, *args, **kwargs)
        state = self._transform_observations(state=state)
        return state


class TransformActionEnvWrapper(EnvironmentWrapper, abc.ABC):
    def __init__(self, env: BaseEnvironment) -> None:
        super().__init__(env=env)

    @property
    @abc.abstractmethod
    def action_space(self) -> SpaceType:
        raise NotImplementedError

    @abc.abstractmethod
    def _transform_action(
        self, action: chex.Array, state: BaseEnvState
    ) -> Tuple[chex.Array, BaseEnvState]:
        raise NotImplementedError

    def step(
        self, state: BaseEnvState, action: chex.Array, *args, **kwargs
    ) -> BaseEnvState:
        action, state = self._transform_action(action=action, state=state)
        state = self._env.step(state=state, action=action, *args, **kwargs)
        return state



from __future__ import annotations

import abc
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import chex
import gymnasium
import mujoco
import numpy as np
from flax import struct
from gymnasium.core import RenderFrame
from mujoco import mjx

import moojoco.environment.mjx_spaces as mjx_spaces
from moojoco.environment.renderer import MujocoRenderer
from moojoco.mjcf.arena import MJCFArena
from moojoco.mjcf.morphology import MJCFMorphology


class MuJoCoEnvironmentConfiguration:
    PHYSICS_TIMESTEP = 0.002

    def __init__(
        self,
        time_scale: float = 1.0,
        num_physics_steps_per_control_step: int = 1,
        simulation_time: float = 10,
        camera_ids: List[int] | None = None,
        render_size: Tuple[int, int] | None = (240, 320),
        render_mode: Optional[str] = None,
        solver_iterations: int | None = None,
        solver_ls_iterations: int | None = None,
        disable_eulerdamp: bool | None = None,
    ) -> None:
        self.time_scale = time_scale
        self.num_physics_steps_per_control_step = num_physics_steps_per_control_step
        self.simulation_time = simulation_time
        self.camera_ids = camera_ids or [0]
        self.render_size = render_size
        self.render_mode = render_mode
        self.solver_iterations = solver_iterations
        self.solver_ls_iterations = solver_ls_iterations
        self.disable_eulerdamp = disable_eulerdamp

    @property
    def control_timestep(self) -> float:
        return self.num_physics_steps_per_control_step * self.physics_timestep

    @property
    def physics_timestep(self) -> float:
        return self.PHYSICS_TIMESTEP * self.time_scale

    @property
    def total_num_control_steps(self) -> int:
        return int(np.ceil(self.simulation_time / self.control_timestep))

    @property
    def total_num_physics_steps(self) -> int:
        return int(np.ceil(self.simulation_time / self.physics_timestep))


ModelType = Union[mujoco.MjModel, mjx.Model]
DataType = Union[mujoco.MjData, mjx.Data]
SpaceType = Union[gymnasium.spaces.Space, mjx_spaces.Space]
BoxSpaceType = Union[gymnasium.spaces.Box, mjx_spaces.Box]
DictSpaceType = Union[gymnasium.spaces.Dict, mjx_spaces.Dict]


class BaseObservable:
    def __init__(
        self,
        name: str,
        low: chex.Array,
        high: chex.Array,
        retriever: Callable[[BaseEnvState], chex.Array],
    ) -> None:
        self.name = name
        self.low = low
        self.high = high
        self.retriever = retriever

    def __call__(self, state: BaseEnvState) -> chex.Array:
        return self.retriever(state)

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.low.shape


@struct.dataclass
class BaseEnvState(abc.ABC):
    mj_model: mujoco.MjModel = struct.field(pytree_node=False)
    mj_data: mujoco.MjData = struct.field(pytree_node=False)
    observations: Dict[str, chex.Array]
    reward: chex.Array
    terminated: chex.Array
    truncated: chex.Array
    info: Dict[str, Any]
    rng: np.random.RandomState | chex.PRNGKey


class BaseEnvironment(abc.ABC):
    def __init__(self, configuration: MuJoCoEnvironmentConfiguration) -> None:
        self._configuration = configuration

    @property
    def environment_configuration(self) -> MuJoCoEnvironmentConfiguration:
        return self._configuration

    @property
    @abc.abstractmethod
    def action_space(self) -> SpaceType:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def actuators(self) -> List[str]:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def observation_space(self) -> SpaceType:
        raise NotImplementedError

    @abc.abstractmethod
    def step(
        self, state: BaseEnvState, action: chex.Array, *args, **kwargs
    ) -> BaseEnvState:
        raise NotImplementedError

    @abc.abstractmethod
    def reset(
        self, rng: np.random.RandomState | chex.PRNGKey, *args, **kwargs
    ) -> BaseEnvState:
        raise NotImplementedError

    @abc.abstractmethod
    def render(self, state: BaseEnvState) -> List[RenderFrame] | None:
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        raise NotImplementedError


class BaseMuJoCoEnvironment(BaseEnvironment, abc.ABC):
    metadata = {"render_modes": []}

    def __init__(
        self,
        mjcf_str: str,
        mjcf_assets: Dict[str, Any],
        configuration: MuJoCoEnvironmentConfiguration,
    ) -> None:
        super().__init__(configuration=configuration)
        self._mjcf_str = mjcf_str
        self._mjcf_assets = mjcf_assets

        self._renderers: Dict[int, Union[MujocoRenderer, mujoco.Renderer]] = dict()

        self._mj_model, self._mj_data = self._initialize_mj_model_and_data()

        assert (
            configuration.render_mode is None
            or self.environment_configuration.render_mode
            in self.metadata["render_modes"]
        ), (
            f"Unsupported render mode: '{self.environment_configuration.render_mode}'. Must be one "
            f"of {self.metadata['render_modes']}"
        )

        self._observables: Optional[List[BaseObservable]] = self._create_observables()
        self._action_space: Optional[BoxSpaceType] = self._create_action_space()
        self._observation_space: Optional[DictSpaceType] = (
            self._create_observation_space()
        )

    @classmethod
    def from_morphology_and_arena(
        cls,
        morphology: MJCFMorphology,
        arena: MJCFArena,
        configuration: MuJoCoEnvironmentConfiguration,
    ) -> BaseMuJoCoEnvironment:
        arena.attach(other=morphology, free_joint=True)
        mjcf_str, mjcf_assets = arena.get_mjcf_str(), arena.get_mjcf_assets()
        return cls(
            mjcf_str=mjcf_str, mjcf_assets=mjcf_assets, configuration=configuration
        )

    @property
    def frozen_mj_model(self) -> mujoco.MjModel:
        return self._mj_model

    @property
    def frozen_mj_data(self) -> mujoco.MjData:
        return self._mj_data

    @property
    def actuators(self) -> List[str]:
        return [
            self.frozen_mj_model.actuator(i).name
            for i in range(self.frozen_mj_model.nu)
        ]

    @property
    def action_space(self) -> SpaceType:
        return self._action_space

    @property
    def observation_space(self) -> SpaceType:
        return self._observation_space

    def _initialize_mj_model_and_data(self) -> Tuple[mujoco.MjModel, mujoco.MjData]:
        mj_model = mujoco.MjModel.from_xml_string(
            xml=self._mjcf_str, assets=self._mjcf_assets
        )
        mj_model.vis.global_.offheight = self.environment_configuration.render_size[0]
        mj_model.vis.global_.offwidth = self.environment_configuration.render_size[1]
        mj_model.opt.timestep = self.environment_configuration.physics_timestep
        if self.environment_configuration.solver_iterations:
            mj_model.opt.iterations = self.environment_configuration.solver_iterations
        if self.environment_configuration.solver_ls_iterations:
            mj_model.opt.ls_iterations = (
                self.environment_configuration.solver_ls_iterations
            )
        if self.environment_configuration.disable_eulerdamp:
            mj_model.opt.disableflags |= mujoco.mjtDisableBit.mjDSBL_EULERDAMP
        mj_data = mujoco.MjData(mj_model)
        return mj_model, mj_data

    def step(
        self, state: BaseEnvState, action: chex.Array, *args, **kwargs
    ) -> BaseEnvState:
        previous_state = state
        state = self._update_simulation(state=state, ctrl=action)

        state = self._update_observations(state=state)
        state = self._update_reward(state=state, previous_state=previous_state)
        state = self._update_terminated(state=state)
        state = self._update_truncated(state=state)
        state = self._update_info(state=state)

        return state

    def get_renderer(
        self,
        identifier: int,
        model: mujoco.MjModel,
        data: mujoco.MjData,
        state: BaseEnvState,
    ) -> Union[MujocoRenderer, mujoco.Renderer]:
        if identifier not in self._renderers:
            if self.environment_configuration.render_mode == "human":
                renderer = MujocoRenderer(
                    model=model, data=data, default_cam_config=None
                )
            else:
                renderer = mujoco.Renderer(
                    model=model,
                    height=self.environment_configuration.render_size[0],
                    width=self.environment_configuration.render_size[1],
                )
            self._renderers[identifier] = renderer
        return self._renderers[identifier]

    def get_renderer_context(
        self, renderer: Union[MujocoRenderer, mujoco.Renderer]
    ) -> mujoco.MjrContext:
        try:
            # noinspection PyProtectedMember
            context = renderer._get_viewer(
                render_mode=self.environment_configuration.render_mode
            ).con
        except AttributeError:
            # noinspection PyProtectedMember
            context = renderer._mjr_context
        return context

    @abc.abstractmethod
    def _get_mj_models_and_datas_to_render(
        self, state: BaseEnvState
    ) -> Tuple[List[mujoco.MjModel], List[mujoco.MjData]]:
        raise NotImplementedError

    def render(self, state: BaseEnvState) -> List[RenderFrame] | None:
        camera_ids = self.environment_configuration.camera_ids or [-1]

        mj_models, mj_datas = self._get_mj_models_and_datas_to_render(state=state)

        frames = []
        for i, (model, data) in enumerate(zip(mj_models, mj_datas)):
            mujoco.mj_forward(m=model, d=data)
            renderer = self.get_renderer(
                identifier=i, model=model, data=data, state=state
            )

            if self.environment_configuration.render_mode == "human":
                viewer = renderer._get_viewer("human")
                viewer.model = model
                viewer.data = data

                return renderer.render(render_mode="human", camera_id=camera_ids[0])
            else:
                for camera_id in camera_ids:
                    renderer._model = model
                    renderer.update_scene(data=data, camera=camera_id)
                    frame = renderer.render()[:, :, ::-1]
                    frames.append(frame)

        return frames

    def _close_renderers(self) -> None:
        for renderer in self._renderers.values():
            renderer.close()

    def close(self) -> None:
        self._close_renderers()
        del self._mj_model
        del self._mj_data
        del self._observation_space
        del self._action_space
        del self._observables

    @abc.abstractmethod
    def _create_observation_space(self) -> DictSpaceType:
        raise NotImplementedError

    @abc.abstractmethod
    def _update_observations(self, state: BaseEnvState) -> BaseEnvState:
        raise NotImplementedError

    @abc.abstractmethod
    def _create_action_space(self) -> BoxSpaceType:
        raise NotImplementedError

    @abc.abstractmethod
    def _update_simulation(self, state: BaseEnvState, ctrl: chex.Array) -> BaseEnvState:
        raise NotImplementedError

    @abc.abstractmethod
    def _prepare_reset(self) -> Tuple[Tuple[ModelType, DataType], ...]:
        raise NotImplementedError

    @abc.abstractmethod
    def _finish_reset(
        self,
        models_and_datas: Tuple[Tuple[ModelType, DataType], ...],
        rng: np.random.RandomState | chex.PRNGKey,
    ) -> BaseEnvState:
        raise NotImplementedError

    @abc.abstractmethod
    def reset(
        self, rng: np.random.RandomState | chex.PRNGKey, *args, **kwargs
    ) -> BaseEnvState:
        raise NotImplementedError

    @abc.abstractmethod
    def _create_observables(self) -> List[BaseObservable]:
        raise NotImplementedError

    @abc.abstractmethod
    def _update_reward(
        self, state: BaseEnvState, previous_state: BaseEnvState
    ) -> BaseEnvState:
        raise NotImplementedError

    @abc.abstractmethod
    def _update_terminated(self, state: BaseEnvState) -> BaseEnvState:
        raise NotImplementedError

    @abc.abstractmethod
    def _update_truncated(self, state: BaseEnvState) -> BaseEnvState:
        raise NotImplementedError

    @abc.abstractmethod
    def _update_info(self, state: BaseEnvState) -> BaseEnvState:
        raise NotImplementedError



from __future__ import annotations

from abc import ABC
from typing import List, Union

import numpy as np
from dm_control import mjcf
from fprs.robot import Morphology
from fprs.specification import MorphologySpecification

from moojoco.mjcf.component import MJCFRootComponent, MJCFSubComponent


class MJCFMorphology(Morphology, MJCFRootComponent, ABC):
    def __init__(
        self,
        specification: MorphologySpecification,
        name: str = "morphology",
        *args,
        **kwargs,
    ) -> None:
        Morphology.__init__(self, specification=specification)
        MJCFRootComponent.__init__(self, name=name, *args, **kwargs)

    @property
    def actuators(self) -> List[mjcf.Element]:
        return self.mjcf_model.find_all("actuator")

    @property
    def sensors(self) -> List[mjcf.Element]:
        return self.mjcf_model.find_all("sensor")

    def _build(self, *args, **kwargs) -> None:
        raise NotImplementedError


class MJCFMorphologyPart(MJCFSubComponent, ABC):
    def __init__(
        self,
        parent: Union[MJCFMorphology, MJCFMorphologyPart],
        name: str,
        position: np.array,
        euler: np.array,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            parent=parent, name=name, position=position, euler=euler, *args, **kwargs
        )

    @property
    def parent(self) -> Union[MJCFMorphology, MJCFMorphologyPart]:
        return self._parent

    @property
    def morphology_specification(self) -> MorphologySpecification:
        return self.parent.morphology_specification



from __future__ import annotations

import abc
from abc import ABC
from typing import Dict, Optional, Union

import numpy as np
from dm_control import mjcf
from dm_control.mjcf import export_with_assets
from dm_control.mjcf.element import _AttachmentFrame
from scipy.spatial.transform import Rotation


class MJCFRootComponent(ABC):
    def __init__(self, name: str, *args, **kwargs) -> None:
        self._name = name
        self._parent = None
        self._mjcf_model = mjcf.RootElement(model=name)
        self._build(*args, **kwargs)

    @property
    def base_name(self) -> str:
        return self._name

    @property
    def mjcf_model(self) -> mjcf.RootElement:
        return self._mjcf_model

    @property
    def mjcf_body(self) -> mjcf.Element:
        return self.mjcf_model.worldbody

    def attach(
        self,
        other: MJCFRootComponent,
        position: Optional[np.ndarray] = None,
        euler: Optional[np.ndarray] = None,
        free_joint: bool = False,
    ) -> _AttachmentFrame:
        attachment_site = self.mjcf_body.add(
            "site",
            name=f"{self.base_name}_attachment_{other.base_name}",
            pos=position,
            euler=euler,
        )
        other._parent = self

        frame = attachment_site.attach(other.mjcf_model)
        if free_joint:
            frame.add("freejoint", name=f"freejoint")
        return frame

    def detach(self) -> None:
        self.mjcf_model.detach()

    def export_to_xml_with_assets(self, output_directory: str = "./mjcf") -> None:
        export_with_assets(mjcf_model=self.mjcf_model, out_dir=output_directory)

    def get_mjcf_str(self) -> str:
        return self.mjcf_model.to_xml_string()

    def get_mjcf_assets(self) -> Dict:
        return self.mjcf_model.get_assets()

    @property
    def world_coordinates(self) -> np.ndarray:
        return np.zeros(3)

    @property
    def coordinate_frame_in_root_frame(self) -> (np.ndarray, np.ndarray):
        return np.zeros(3), Rotation.from_euler("xyz", [0, 0, 0]).as_matrix()

    def coordinates_of_point_in_root_frame(self, point: np.ndarray) -> np.ndarray:
        return point

    @abc.abstractmethod
    def _build(self, *args, **kwargs) -> None:
        raise NotImplementedError


class MJCFSubComponent(ABC):
    def __init__(
        self,
        parent: Union[MJCFSubComponent, MJCFRootComponent],
        name: str,
        position: Optional[np.array] = None,
        euler: Optional[np.array] = None,
        *args,
        **kwargs,
    ) -> None:
        self._parent = parent
        self._name = name
        self._mjcf_body = parent.mjcf_body.add(
            "body", name=self._name, pos=position, euler=euler
        )
        self._coordinate_frame_in_root_frame = None
        self._build(*args, **kwargs)

    @property
    def mjcf_model(self) -> mjcf.RootElement:
        return self._parent.mjcf_model

    @property
    def mjcf_body(self) -> mjcf.Element:
        return self._mjcf_body

    @property
    def base_name(self) -> str:
        return self._name

    def attach(
        self,
        other: MJCFRootComponent,
        position: Optional[np.ndarray] = None,
        euler: Optional[np.ndarray] = None,
    ) -> None:
        attachment_site = self.mjcf_body.add(
            "site",
            name=f"{self.base_name}_attachment_{other.base_name}",
            pos=position,
            euler=euler,
        )
        other._parent = self
        frame = attachment_site.attach(other.mjcf_model)
        return frame

    @property
    def coordinate_frame_in_root_frame(self) -> (np.ndarray, np.ndarray):
        """
        Returns the MJCFSubComponent's coordinate frame with respect to the MJCFRootComponent's frame:
            0 -> the object's coordinates in root's frame
            1 -> a rotation matrix that represents the transformation from the object's local frame to the root's frame
        :return:
        """
        if self._coordinate_frame_in_root_frame is None:
            parent_origin, parent_rot = self._parent.coordinate_frame_in_root_frame

            pos = self.mjcf_body.pos
            euler = self.mjcf_body.euler
            my_rot = Rotation.from_euler("xyz", euler).as_matrix()

            self._coordinate_frame_in_root_frame = parent_origin + parent_rot.dot(
                pos
            ), parent_rot.dot(my_rot)
        return self._coordinate_frame_in_root_frame

    def coordinates_of_point_in_root_frame(self, point: np.ndarray) -> np.ndarray:
        """Returns the world coordinates of the given point in local coordinates"""
        origin, rot = self.coordinate_frame_in_root_frame
        return origin + rot.dot(point)

    @property
    def world_coordinates(self) -> np.ndarray:
        return self.coordinate_frame_in_root_frame[0]

    @abc.abstractmethod
    def _build(self, *args, **kwargs) -> None:
        raise NotImplementedError



from __future__ import annotations

import abc
import copy
from abc import ABC
from typing import Any, Callable, Dict, List, Tuple

import chex
import jax
import jax.numpy as jnp
import mujoco
import numpy as np
from flax import struct
from mujoco import mjx

import moojoco.environment.mjx_spaces as mjx_spaces
from moojoco.environment.base import (
    BaseEnvState,
    BaseMuJoCoEnvironment,
    BaseObservable,
    MuJoCoEnvironmentConfiguration,
)
from moojoco.mjcf.arena import MJCFArena
from moojoco.mjcf.morphology import MJCFMorphology


def mjx_get_model(
    mj_model: mujoco.MjModel, mjx_model: mjx.Model, n_mj_models: int = 1
) -> List[mujoco.MjModel]:
    """
    Transfer mjx.Model to mujoco.MjModel
    """
    mj_models = [copy.deepcopy(mj_model) for _ in range(n_mj_models)]

    offloaded_mjx_model = jax.device_get(mjx_model)
    for key, v in vars(offloaded_mjx_model).items():
        try:
            for i, model in enumerate(mj_models):
                previous_value = getattr(model, key)
                if isinstance(previous_value, np.ndarray):
                    if previous_value.shape != v.shape:
                        actual_value = v[i]
                    else:
                        actual_value = v
                    previous_value[:] = actual_value
                else:
                    setattr(model, key, v)
        except AttributeError:
            pass
        except ValueError:
            pass
        except IndexError:
            pass
    return mj_models


@struct.dataclass
class MJXEnvState(BaseEnvState):
    mjx_model: mjx.Model
    mjx_data: mjx.Data
    observations: Dict[str, jax.Array]
    rng: chex.PRNGKey


class MJXObservable(BaseObservable):
    def __init__(
        self,
        name: str,
        low: jnp.ndarray,
        high: jnp.ndarray,
        retriever: Callable[[MJXEnvState], jnp.ndarray],
    ) -> None:
        super().__init__(name=name, low=low, high=high, retriever=retriever)

    def __call__(self, state: MJXEnvState) -> jnp.ndarray:
        return super().__call__(state=state)


class MJXEnv(BaseMuJoCoEnvironment, ABC):
    def __init__(
        self,
        mjcf_str: str,
        mjcf_assets: Dict[str, Any],
        configuration: MuJoCoEnvironmentConfiguration,
    ) -> None:
        super().__init__(
            mjcf_str=mjcf_str, mjcf_assets=mjcf_assets, configuration=configuration
        )
        self._mjx_model, self._mjx_data = self._initialize_mjx_model_and_data()

    @classmethod
    def from_morphology_and_arena(
        cls,
        morphology: MJCFMorphology,
        arena: MJCFArena,
        configuration: MuJoCoEnvironmentConfiguration,
    ) -> MJXEnv:
        return super().from_morphology_and_arena(
            morphology=morphology, arena=arena, configuration=configuration
        )

    @property
    def frozen_mjx_model(self) -> mjx.Model:
        return self._mjx_model

    @property
    def frozen_mjx_data(self) -> mjx.Data:
        return self._mjx_data

    def _initialize_mjx_model_and_data(self) -> Tuple[mjx.Model, mjx.Data]:
        return mjx.put_model(m=self.frozen_mj_model), mjx.put_data(
            m=self.frozen_mj_model, d=self.frozen_mj_data
        )

    @staticmethod
    def _get_batch_size(state: MJXEnvState) -> int:
        try:
            return state.reward.shape[0]
        except IndexError:
            return 1
        except AttributeError:
            return 1

    def _get_mj_models_and_datas_to_render(
        self, state: MJXEnvState
    ) -> Tuple[List[mujoco.MjModel], List[mujoco.MjData]]:
        num_models = (
            1
            if self.environment_configuration.render_mode == "human"
            else self._get_batch_size(state=state)
        )
        mj_models = mjx_get_model(
            mj_model=state.mj_model, mjx_model=state.mjx_model, n_mj_models=num_models
        )
        mj_datas = mjx.get_data(m=mj_models[0], d=state.mjx_data)
        if not isinstance(mj_datas, list):
            mj_datas = [mj_datas]
        return mj_models, mj_datas

    @property
    def observables(self) -> List[MJXObservable]:
        return self._observables

    def _create_observation_space(self) -> mjx_spaces.Dict:
        observation_space = dict()
        for observable in self.observables:
            observation_space[observable.name] = mjx_spaces.Box(
                low=observable.low, high=observable.high, shape=observable.shape
            )
        return mjx_spaces.Dict(observation_space)

    def _create_action_space(self) -> mjx_spaces.Box:
        bounds = jnp.array(
            self.frozen_mj_model.actuator_ctrlrange.copy().astype(np.float32)
        )
        low, high = bounds.T
        action_space = mjx_spaces.Box(
            low=low, high=high, shape=low.shape, dtype=jnp.float32
        )
        return action_space

    def _update_observations(self, state: MJXEnvState) -> MJXEnvState:
        observations = jax.tree_util.tree_map(
            lambda observable: (observable.name, observable(state=state)),
            self.observables,
        )
        # noinspection PyUnresolvedReferences
        return state.replace(observations=dict(observations))

    def _update_simulation(self, state: MJXEnvState, ctrl: jnp.ndarray) -> MJXEnvState:
        mjx_data = state.mjx_data

        def _simulation_step(_data, _) -> Tuple[mjx.Data, None]:
            _data = _data.replace(ctrl=ctrl)
            return mjx.step(state.mjx_model, _data), None

        mjx_data, _ = jax.lax.scan(
            _simulation_step,
            mjx_data,
            (),
            self.environment_configuration.num_physics_steps_per_control_step,
        )
        # noinspection PyUnresolvedReferences
        return state.replace(mjx_data=mjx_data)

    def close(self) -> None:
        super().close()
        del self._mjx_model
        del self._mjx_data

    def _prepare_reset(
        self,
    ) -> Tuple[Tuple[mujoco.MjModel, mujoco.MjData], Tuple[mjx.Model, mjx.Data]]:
        mj_model = self.frozen_mj_model
        mj_data = self.frozen_mj_data

        mjx_model = self.frozen_mjx_model
        mjx_data = self.frozen_mjx_data
        return (mj_model, mj_data), (mjx_model, mjx_data)

    def _finish_reset(
        self,
        models_and_datas: Tuple[
            Tuple[mujoco.MjModel, mujoco.MjData], Tuple[mjx.Model, mjx.Data]
        ],
        rng: np.random.RandomState,
    ) -> MJXEnvState:
        (mj_model, mj_data), (mjx_model, mjx_data) = models_and_datas
        mjx_data = mjx.forward(m=mjx_model, d=mjx_data)
        state = MJXEnvState(
            mj_model=mj_model,
            mj_data=mj_data,
            mjx_model=mjx_model,
            mjx_data=mjx_data,
            observations={},
            reward=0,
            terminated=False,
            truncated=False,
            info={},
            rng=rng,
        )
        state = self._update_observations(state=state)
        state = self._update_info(state=state)
        return state

    @abc.abstractmethod
    def _create_observables(self) -> List[MJXObservable]:
        raise NotImplementedError

    @abc.abstractmethod
    def reset(self, rng: jnp.ndarray, *args, **kwargs) -> MJXEnvState:
        raise NotImplementedError

    @abc.abstractmethod
    def _update_reward(
        self, state: MJXEnvState, previous_state: MJXEnvState
    ) -> MJXEnvState:
        raise NotImplementedError

    @abc.abstractmethod
    def _update_terminated(self, state: MJXEnvState) -> MJXEnvState:
        raise NotImplementedError

    @abc.abstractmethod
    def _update_truncated(self, state: MJXEnvState) -> MJXEnvState:
        raise NotImplementedError

    @abc.abstractmethod
    def _update_info(self, state: MJXEnvState) -> MJXEnvState:
        raise NotImplementedError



"""
Taken from https://github.com/RobertTLange/gymnax, with some slight modifications:
- gym -> gymnasium
- types:
    - Accept array types for Box.low and Box.high as well
- Added __repr__
"""

from collections import OrderedDict
from typing import Any, Dict, Sequence, Tuple

import chex
import jax
import jax.numpy as jnp
import numpy as np
from gymnasium import spaces as gspc


def _short_repr(arr: jnp.ndarray) -> str:
    if arr.size != 0 and jnp.min(arr) == jnp.max(arr):
        return str(jnp.min(arr))
    return str(arr)


class Space:
    """
    Minimal jittable class for abstract gymnax space.
    """

    def sample(self, rng: chex.PRNGKey) -> chex.Array:
        raise NotImplementedError

    def contains(self, x: jnp.int_) -> bool:
        raise NotImplementedError


class Discrete(Space):
    """
    Minimal jittable class for discrete gymnax spaces.
    TODO: For now this is a 1d space. Make composable for multi-discrete.
    """

    def __init__(self, num_categories: int):
        assert num_categories >= 0
        self.n = num_categories
        self.shape = ()
        self.dtype = jnp.int_

    def sample(self, rng: chex.PRNGKey) -> chex.Array:
        """Sample random action uniformly from set of categorical choices."""
        return jax.random.randint(
            rng, shape=self.shape, minval=0, maxval=self.n
        ).astype(self.dtype)

    def contains(self, x: jnp.int_) -> bool:
        """Check whether specific object is within space."""
        # type_cond = isinstance(x, self.dtype)
        # shape_cond = (x.shape == self.shape)
        range_cond = jnp.logical_and(x >= 0, x < self.n)
        return range_cond


class Box(Space):
    """
    Minimal jittable class for array-shaped gymnax spaces.
    TODO: Add unboundedness - sampling from other distributions, etc.
    """

    def __init__(
        self,
        low: float | jax.Array,
        high: float | jax.Array,
        shape: Tuple[int],
        dtype: jnp.dtype = jnp.float32,
    ):
        self.low = low
        self.high = high
        self.shape = shape
        self.dtype = dtype

    def sample(self, rng: chex.PRNGKey) -> chex.Array:
        """Sample random action uniformly from 1D continuous range."""
        return jax.random.uniform(
            rng, shape=self.shape, minval=self.low, maxval=self.high
        ).astype(self.dtype)

    def contains(self, x: jnp.int_) -> bool:
        """Check whether specific object is within space."""
        # type_cond = isinstance(x, self.dtype)
        # shape_cond = (x.shape == self.shape)
        range_cond = jnp.logical_and(jnp.all(x >= self.low), jnp.all(x <= self.high))
        return range_cond

    def __repr__(self) -> str:
        """A string representation of this space.

        The representation will include bounds, shape and dtype.
        If a bound is uniform, only the corresponding scalar will be given to avoid redundant and ugly strings.

        Returns:
            A representation of the space
        """
        return f"Box({_short_repr(self.low)}, {_short_repr(self.high)}, {self.shape}, {self.dtype})"


class Dict(Space):
    """Minimal jittable class for dictionary of simpler jittable spaces."""

    def __init__(self, spaces: Dict[Any, Space]):
        self.spaces = spaces
        self.num_spaces = len(spaces)

    def sample(self, rng: chex.PRNGKey) -> Dict:
        """Sample random action from all subspaces."""
        key_split = jax.random.split(rng, self.num_spaces)
        return OrderedDict(
            [
                (k, self.spaces[k].sample(key_split[i]))
                for i, k in enumerate(self.spaces)
            ]
        )

    def contains(self, x: jnp.int_) -> bool:
        """Check whether dimensions of object are within subspace."""
        # type_cond = isinstance(x, Dict)
        # num_space_cond = len(x) != len(self.spaces)
        # Check for each space individually
        out_of_space = 0
        for k, space in self.spaces.items():
            out_of_space += 1 - space.contains(getattr(x, k))
        return out_of_space == 0

    def __repr__(self) -> str:
        """Gives a string representation of this space."""
        return (
            "Dict(" + ", ".join([f"{k!r}: {s}" for k, s in self.spaces.items()]) + ")"
        )


class Tuple(Space):
    """Minimal jittable class for tuple (product) of jittable spaces."""

    def __init__(self, spaces: Sequence[Space]):
        self.spaces = spaces
        self.num_spaces = len(spaces)

    def sample(self, rng: chex.PRNGKey) -> Tuple[chex.Array]:
        """Sample random action from all subspaces."""
        key_split = jax.random.split(rng, self.num_spaces)
        return tuple([s.sample(key_split[i]) for i, s in enumerate(self.spaces)])

    def contains(self, x: jnp.int_) -> bool:
        """Check whether dimensions of object are within subspace."""
        # type_cond = isinstance(x, tuple)
        # num_space_cond = len(x) != len(self.spaces)
        # Check for each space individually
        out_of_space = 0
        for i, space in enumerate(self.spaces):
            out_of_space += 1 - space.contains(x[i])
        return out_of_space == 0

    def __repr__(self) -> str:
        """Gives a string representation of this space."""
        return "Tuple(" + ", ".join([str(s) for s in self.spaces]) + ")"


def gymnax_space_to_gym_space(space: Space) -> gspc.Space:
    """Convert Gymnax space to equivalent Gym space"""
    if isinstance(space, Discrete):
        return gspc.Discrete(space.n)
    elif isinstance(space, Box):
        low = (
            float(space.low)
            if (np.isscalar(space.low) or space.low.size == 1)
            else np.array(space.low)
        )
        high = (
            float(space.high)
            if (np.isscalar(space.high) or space.low.size == 1)
            else np.array(space.high)
        )
        return gspc.Box(low, high, space.shape, space.dtype)
    elif isinstance(space, Dict):
        return gspc.Dict({k: gymnax_space_to_gym_space(v) for k, v in space.spaces})
    elif isinstance(space, Tuple):
        return gspc.Tuple(space.spaces)
    else:
        raise NotImplementedError(
            f"Conversion of {space.__class__.__name__} not supported"
        )



from abc import ABC

from moojoco.mjcf.component import MJCFRootComponent


class ArenaConfiguration(ABC):
    def __init__(self, name: str) -> None:
        self.name = name


class MJCFArena(MJCFRootComponent, ABC):
    def __init__(self, configuration: ArenaConfiguration, *args, **kwargs) -> None:
        self._configuration = configuration
        super().__init__(name=configuration.name, *args, **kwargs)

    @property
    def arena_configuration(self) -> ArenaConfiguration:
        return self._configuration

    def _build(self, *args, **kwargs) -> None:
        raise NotImplementedError
